/* BDF - Copyright (c) 2022 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.servlets;

import fr.exemole.bdfserver.api.BdfServer;
import fr.exemole.bdfserver.api.instruction.BdfInstruction;
import fr.exemole.bdfserver.api.instruction.BdfInstructionConstants;
import fr.exemole.bdfserver.api.interaction.Domains;
import fr.exemole.bdfserver.api.session.LoginException;
import fr.exemole.bdfserver.api.users.BdfUser;
import fr.exemole.bdfserver.conf.ConfUtils;
import fr.exemole.bdfserver.multi.MultiConf;
import fr.exemole.bdfserver.multi.MultiResponseHandlerFactory;
import fr.exemole.bdfserver.multi.api.Multi;
import fr.exemole.bdfserver.multi.api.MultiConstants;
import fr.exemole.bdfserver.multi.api.namespaces.IndexSpace;
import fr.exemole.bdfserver.multi.tools.CrossInfo;
import fr.exemole.bdfserver.tools.apps.AppConf;
import fr.exemole.bdfserver.tools.apps.AppConfUtils;
import fr.exemole.bdfserver.tools.instruction.BdfInstructionUtils;
import java.io.IOException;
import java.text.ParseException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import net.fichotheque.SubsetKey;
import net.mapeadores.util.attr.Attribute;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.LangPreference;
import net.mapeadores.util.localisation.LangPreferenceBuilder;
import net.mapeadores.util.logging.CommandMessage;
import net.mapeadores.util.logging.ErrorMessageException;
import net.mapeadores.util.request.OutputInfo;
import net.mapeadores.util.request.RequestConstants;
import net.mapeadores.util.request.RequestUtils;
import net.mapeadores.util.servlets.HttpAccessException;
import net.mapeadores.util.servlets.HttpServletRequestMap;
import net.mapeadores.util.servlets.ResponseHandler;
import net.mapeadores.util.servlets.exceptions.InternalErrorException;
import net.mapeadores.util.servlets.exceptions.NotFoundException;


/**
 *
 * @author Vincent Calame
 */
public final class BDFRoutes {


    private BDFRoutes() {

    }

    public static String getFichothequeRedirect(Multi multi) {
        String value = null;
        Attribute attribute = multi.getMultiMetadata().getAttributes().getAttribute(IndexSpace.REDIRECT_KEY);
        if (attribute != null) {
            value = attribute.getFirstValue();
        } else {
            value = multi.getMultiConf().getParam("default_fichotheque");
        }
        if (value != null) {
            if (!ConfUtils.isValidMultiName(value)) {
                value = null;
            }
        }
        return value;
    }

    public static void resolveMulti(BDFWebapps bdfWebapps, String pathInfo, HttpServletRequestMap requestMap, HttpServletResponse response) throws IOException, ServletException {
        Multi multi = bdfWebapps.getMulti();
        int idx = pathInfo.indexOf('/');
        if (idx == -1) {
            resolveMultiRoute(bdfWebapps, checkAlias(pathInfo), requestMap, response);
        } else {
            String pathRoot = pathInfo.substring(0, idx);
            String relativePath = pathInfo.substring(idx + 1);
            switch (pathRoot) {
                case "multi-rsc":
                case "_rsc": {
                    ResponseHandler responseHandler = MultiResponseHandlerFactory.getResourceHandler(multi, relativePath);
                    if (responseHandler != null) {
                        responseHandler.handleResponse(response);
                    } else {
                        throw new NotFoundException();
                    }
                    break;
                }
                default: {
                    BdfServer bdfServer;
                    try {
                        bdfServer = multi.getBdfServer(pathRoot);
                    } catch (ErrorMessageException eme) {
                        CommandMessage errorMessage = eme.getErrorMessage();
                        switch (errorMessage.getMessageKey()) {
                            case MultiConstants.UNKNOWN_FICHOTHEQUE_ERROR:
                                throw new NotFoundException(errorMessage);
                            case MultiConstants.INACTIVE_FICHOTHEQUE_ERROR:
                                throw new HttpAccessException(HttpServletResponse.SC_SERVICE_UNAVAILABLE, errorMessage);
                            default:
                                throw new InternalErrorException(errorMessage);
                        }
                    }
                    if (relativePath.isEmpty()) {
                        response.sendRedirect(Domains.SESSION);
                    } else {
                        BDFRoutes.resolveBdfServer(bdfWebapps, bdfServer, relativePath, requestMap, response);
                    }
                }
            }
        }
    }

    private static void resolveMultiRoute(BDFWebapps bdfWebapps, String route, HttpServletRequestMap requestMap, HttpServletResponse response) throws IOException, ServletException {
        Multi multi = bdfWebapps.getMulti();
        int idx2 = route.indexOf('-');
        if (idx2 == -1) {
            if (multi.containsFichotheque(route)) {
                response.sendRedirect(route + "/" + Domains.SESSION);
            } else {
                throw new NotFoundException();
            }
        } else {
            String firstPart = route.substring(0, idx2);
            String secondPart = route.substring(idx2 + 1);
            ResponseHandler responseHandler;
            switch (firstPart) {
                case "multi": {
                    switch (secondPart) {
                        case "admin":
                            responseHandler = resolveAdminMulti(bdfWebapps, requestMap);
                            break;
                        case "run":
                            responseHandler = resolveRunMulti(bdfWebapps, requestMap);
                            break;
                        default:
                            throw new NotFoundException();
                    }
                    break;
                }
                case "app": {
                    AppConf appConf = AppConfUtils.getAppConf(multi.getWebappsResourceStorages(), secondPart);
                    if ((appConf != null) && (appConf.getBoolean(AppConf.MULTI_ACTIVE))) {
                        String crossLogin = appConf.getString(AppConf.MULTI_CROSSLOGIN);
                        CrossInfo crossInfo = null;
                        if ((crossLogin != null) && (multi.isAuthentificationSharingAllowed())) {
                            SubsetKey sphereKey = getValidSphere(multi, crossLogin);
                            if (sphereKey != null) {
                                crossInfo = CrossInfoFactory.newInstance(bdfWebapps, sphereKey, requestMap);
                            } else {
                                throw new InternalErrorException("_ error.wrong.multi_crosslogin");
                            }
                        }
                        responseHandler = MultiResponseHandlerFactory.getAppHandler(multi, appConf, crossInfo, requestMap);
                    } else {
                        throw new NotFoundException();
                    }
                    break;
                }
                case "index": {
                    if (!multi.isAuthentificationSharingAllowed()) {
                        throw new NotFoundException();
                    }
                    SubsetKey sphereKey = getValidSphere(multi, secondPart);
                    if (sphereKey != null) {
                        responseHandler = resolveSphereIndex(bdfWebapps, sphereKey, requestMap);
                    } else {
                        throw new NotFoundException();
                    }
                    break;
                }
                default:
                    throw new NotFoundException();
            }
            responseHandler.handleResponse(response);
        }
    }

    public static void resolveBdfServer(BDFWebapps bdfWebapps, BdfServer bdfServer, String pathInfo, HttpServletRequestMap requestMap, HttpServletResponse response) throws IOException, ServletException {
        HttpServletRequest request = requestMap.getRequest();
        if (bdfServer.getInitState() == BdfServer.NOADMIN_STATE) {
            throw new InternalErrorException("_ error.unsupported.noadminstate");
        }
        BdfInstruction bdfInstruction = BDFUtils.getBdfInstruction(bdfServer, requestMap, pathInfo);
        if (bdfInstruction == null) {
            throw new NotFoundException();
        }
        BdfUser bdfUser = null;
        short needBdfUser = bdfInstruction.getBdfUserNeed();
        if (needBdfUser != BdfInstructionConstants.NONE_BDFUSER) {
            try {
                bdfUser = BDFUtils.getCurrentBdfUser(bdfServer, request);
                if (needBdfUser == BdfInstructionConstants.EXIT_BDFUSER) {
                    if (bdfUser != null) {
                        boolean all = false;
                        if ((bdfWebapps.isMultiBdf()) && (bdfWebapps.getMulti().isAuthentificationSharingAllowed())) {
                            all = true;
                        }
                        bdfWebapps.invalidateBdfUser(bdfServer, bdfUser, all);
                    }
                    throw new LoginException(LoginException.EXIT_ERROR);
                } else if (bdfUser == null) {
                    bdfUser = bdfWebapps.getMatchingBdfUser(bdfServer, requestMap);
                    if (bdfUser == null) {
                        bdfUser = bdfWebapps.createBdfUser(bdfServer, requestMap);
                    }
                }
            } catch (LoginException le) {
                if (needBdfUser != BdfInstructionConstants.IF_ANY_BDFUSER) {
                    ResponseHandler responseHandler = LoginEngine.resolveLoginException(bdfServer, requestMap, pathInfo, bdfInstruction, le);
                    responseHandler.handleResponse(response);
                    return;
                }
            }
        }
        try {
            ResponseHandler responseHandler = bdfInstruction.runInstruction(bdfUser);
            if (responseHandler == null) {
                throw new NotFoundException();
            }
            long lastModified = responseHandler.getLastModified();
            if (lastModified == ResponseHandler.LASTMODIFIED_NOCACHE) {
                response.setHeader("Cache-Control", "no-store");
            } else if (lastModified > 0) {
                response.setDateHeader("Last-Modified", lastModified);
                try {
                    long since = request.getDateHeader("If-Modified-Since");
                    if (since >= lastModified) {
                        response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
                        return;
                    }
                } catch (IllegalArgumentException iae) {
                }
            }
            responseHandler.handleResponse(response);
        } catch (HttpAccessException hae) {
            if (hae.isUnresolvedMessage()) {
                hae.resolveMessage(BdfInstructionUtils.getMessageLocalisation(requestMap, bdfServer, bdfUser));
            }
            throw hae;
        }
    }

    private static ResponseHandler resolveAdminMulti(BDFWebapps bdfWebapps, HttpServletRequestMap requestMap) throws IOException, ServletException {
        OutputInfo outputInfo = OutputInfo.buildFromRequest(requestMap, "admin");
        Multi multi = bdfWebapps.getMulti();
        MultiConf multiConf = multi.getMultiConf();
        if (outputInfo.getType() == OutputInfo.PAGE_TYPE) {
            multiConf.update();
            if (requestMap.isTrue(MultiConstants.RELOAD_PARAMNAME)) {
                multi.getWebappsJsAnalyser().clearCache();
            }
        }
        ResponseHandler errorResponseHandler = MultiResponseHandlerFactory.checkInitErrorHandler(multi, outputInfo, requestMap);
        if (errorResponseHandler != null) {
            return errorResponseHandler;
        }
        short check = bdfWebapps.checkMultiAdmin(requestMap);
        return MultiResponseHandlerFactory.getAdminHandler(multi, check, requestMap, outputInfo);
    }

    private static ResponseHandler resolveRunMulti(BDFWebapps bdfWebapps, HttpServletRequestMap requestMap) throws IOException, ServletException {
        OutputInfo outputInfo = new OutputInfo(OutputInfo.STREAM_TYPE, "");
        Multi multi = bdfWebapps.getMulti();
        ResponseHandler errorResponseHandler = MultiResponseHandlerFactory.checkInitErrorHandler(multi, outputInfo, requestMap);
        if (errorResponseHandler != null) {
            return errorResponseHandler;
        }
        return MultiResponseHandlerFactory.getRunHandler(multi, requestMap);
    }

    private static ResponseHandler resolveSphereIndex(BDFWebapps bdfWebapps, SubsetKey sphereKey, HttpServletRequestMap requestMap) throws IOException, ServletException {
        OutputInfo outputInfo = OutputInfo.buildFromRequest(requestMap, sphereKey.getSubsetName());
        Multi multi = bdfWebapps.getMulti();
        ResponseHandler errorResponseHandler = MultiResponseHandlerFactory.checkInitErrorHandler(multi, outputInfo, requestMap);
        if (errorResponseHandler != null) {
            return errorResponseHandler;
        }
        CrossInfo crossInfo = CrossInfoFactory.newInstance(bdfWebapps, sphereKey, requestMap);
        String cmd = requestMap.getParameter(RequestConstants.COMMAND_PARAMETER);
        if ((cmd != null) && (cmd.equals("exit"))) {
            if (crossInfo.hasBdfUser()) {
                try {
                    BdfServer bdfServer = multi.getBdfServer(crossInfo.getReferenceFichothequeName());
                    bdfWebapps.invalidateBdfUser(bdfServer, crossInfo.getBdfUser(), true);
                    crossInfo = crossInfo.cloneWithoutBdfUser();
                } catch (ErrorMessageException eme) {

                }
            }
        }
        return MultiResponseHandlerFactory.getSphereIndexHandler(multi, crossInfo, outputInfo, requestMap);
    }

    private static SubsetKey getValidSphere(Multi multi, String secondPart) {
        try {
            SubsetKey sphereKey = SubsetKey.parse(SubsetKey.CATEGORY_SPHERE, secondPart);
            Attribute attribute = multi.getMultiMetadata().getAttributes().getAttribute(IndexSpace.SPHERES_KEY);
            if (attribute != null) {
                if (attribute.contains(sphereKey.getSubsetName())) {
                    return sphereKey;
                }
            }
            return null;
        } catch (ParseException pe) {
            return null;
        }

    }

    public static LangPreference getLangPreference(Multi multi, HttpServletRequestMap requestMap) {
        LangPreferenceBuilder langPreferenceBuilder = new LangPreferenceBuilder();
        RequestUtils.checkLangPreference(requestMap, langPreferenceBuilder);
        langPreferenceBuilder.addLangs(multi.getMultiMetadata().getWorkingLangs());
        langPreferenceBuilder.addLang(Lang.build("fr"));
        return langPreferenceBuilder.toLangPreference();
    }

    private static String checkAlias(String route) {
        switch (route) {
            case "_admin":
            case "admin-multi":
                return "multi-admin";
            default:
                return route;
        }
    }

}
