/* BDF - Copyright (c) 2015-2022 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.servlets;

import fr.exemole.bdfserver.api.BdfExtensionInitializer;
import fr.exemole.bdfserver.api.BdfServer;
import fr.exemole.bdfserver.api.BdfServerConstants;
import fr.exemole.bdfserver.api.externalsource.FichothequeSharing;
import fr.exemole.bdfserver.api.interaction.InteractionConstants;
import fr.exemole.bdfserver.api.session.LoginException;
import fr.exemole.bdfserver.api.session.SessionObserver;
import fr.exemole.bdfserver.api.users.BdfUser;
import fr.exemole.bdfserver.api.users.BdfUserConstants;
import fr.exemole.bdfserver.conf.BdfServerDirs;
import fr.exemole.bdfserver.conf.Conf;
import fr.exemole.bdfserver.conf.ConfConstants;
import fr.exemole.bdfserver.conf.ConfDirs;
import fr.exemole.bdfserver.conf.ConfUtils;
import fr.exemole.bdfserver.email.SmtpManager;
import fr.exemole.bdfserver.impl.BdfServerImpl;
import fr.exemole.bdfserver.multi.MultiConf;
import fr.exemole.bdfserver.multi.api.Multi;
import fr.exemole.bdfserver.multi.api.MultiConstants;
import fr.exemole.bdfserver.multi.api.MultiEditor;
import fr.exemole.bdfserver.multi.api.MultiMetadata;
import fr.exemole.bdfserver.multi.api.MultiMetadataEditor;
import fr.exemole.bdfserver.multi.api.namespaces.FichothequeSpace;
import fr.exemole.bdfserver.multi.commands.CreationParameters;
import fr.exemole.bdfserver.multi.tools.DefaultMultiMetadata;
import fr.exemole.bdfserver.storage.directory.bdfdata.BdfdataDirectories;
import fr.exemole.bdfserver.storage.directory.tools.SphereCopyEngine;
import fr.exemole.bdfserver.tools.BdfMessageLocalisationFactory;
import fr.exemole.bdfserver.tools.storage.ResourceJsAnalyser;
import fr.exemole.bdfserver.tools.storage.Storages;
import fr.exemole.bdfserver.tools.zip.BdfServerZip;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.zip.Deflater;
import java.util.zip.ZipOutputStream;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpSessionBindingEvent;
import javax.servlet.http.HttpSessionBindingListener;
import net.fichotheque.EditOrigin;
import net.fichotheque.Fichotheque;
import net.fichotheque.sphere.Redacteur;
import net.fichotheque.utils.SphereUtils;
import net.mapeadores.util.attr.Attribute;
import net.mapeadores.util.buildinfo.BuildInfo;
import net.mapeadores.util.exceptions.NestedIOException;
import net.mapeadores.util.io.ResourceStorages;
import net.mapeadores.util.jslib.JsAnalyser;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.LangPreference;
import net.mapeadores.util.localisation.MessageLocalisation;
import net.mapeadores.util.localisation.MessageLocalisationProvider;
import net.mapeadores.util.logging.ErrorMessageException;
import net.mapeadores.util.mimetype.MimeTypeResolver;
import net.mapeadores.util.mimetype.MimeTypeUtils;
import net.mapeadores.util.servlets.HttpServletRequestMap;
import net.mapeadores.util.servlets.ServletUtils;


/**
 *
 * @author Vincent Calame
 */
public class BDFWebapps {

    public final static String BDFWEBAPPS_KEY = "fr.exemole.BDFWebapps";
    public final static String BDFSESSION_KEY = "fr.exemole.BDFSession";
    public final static String BDFBINDING_KEY = "fr.exemole.BDFBinding";
    private final Conf conf;
    private final boolean loginPasswordDisabled;
    private final Map<String, BDFSession> sessionsMap = new LinkedHashMap<String, BDFSession>();
    private final BDFBinding bdfBinding = new BDFBinding();
    private final int maxInactiveInterval = 18000;
    private final InternalMulti internalMulti;
    private final Unique unique;

    BDFWebapps(String servletContextName, Conf conf, SmtpManager smtpManager, boolean loginPasswordDisabled, boolean authentificationSharingAllowed) {
        this.conf = conf;
        this.loginPasswordDisabled = loginPasswordDisabled;
        List<BdfExtensionInitializer> extensionInitializerList = BDFInit.extensionInitializerList(conf);
        ResourceStorages webappsResourceStorages = BDFInit.initWebappsResourceStorages(conf, extensionInitializerList);
        BuildInfo buildInfo = BDFInit.buildInfo();
        MimeTypeResolver mimeTypeResolver = MimeTypeUtils.DEFAULT_RESOLVER;
        if (conf.isMultiBdf()) {
            MultiConf multiConf = MultiConf.build(ConfUtils.getMultiIniFile(conf));
            internalMulti = new InternalMulti(multiConf, servletContextName, ConfUtils.getExistingNameSet(conf), webappsResourceStorages, authentificationSharingAllowed, buildInfo, smtpManager, mimeTypeResolver, extensionInitializerList, conf.getMultiDirs());
            unique = null;
        } else {
            internalMulti = null;
            unique = new Unique(webappsResourceStorages, buildInfo, smtpManager, mimeTypeResolver, extensionInitializerList);
        }
    }

    public boolean isMultiBdf() {
        return conf.isMultiBdf();
    }

    public Conf getConf() {
        return conf;
    }

    public Multi getMulti() {
        return internalMulti;
    }

    public Unique getUnique() {
        return unique;
    }

    public boolean isLoginPasswordDisabled() {
        return loginPasswordDisabled;
    }

    public short checkMultiAdmin(HttpServletRequestMap requestMap) {
        HttpServletRequest request = requestMap.getRequest();
        HttpSession httpSession = request.getSession(false);
        if (httpSession != null) {
            BDFSession bdfSession = (BDFSession) httpSession.getAttribute(BDFSESSION_KEY);
            if (bdfSession != null) {
                if (bdfSession.isMultiAdmin()) {
                    return MultiConf.CONNECTED_STATE;
                }
            }
        }
        short check = internalMulti.getMultiConf().checkAuthentification(requestMap);
        if (check == MultiConf.CONNECTED_STATE) {
            if (httpSession == null) {
                httpSession = request.getSession();
                httpSession.setMaxInactiveInterval(maxInactiveInterval);
            }
            BDFSession bdfSession = getOrCreateBDFSession(httpSession);
            bdfSession.setMultiAdmin(true);
        }
        return check;
    }

    public BdfUser getMatchingBdfUser(BdfServer bdfServer, HttpServletRequestMap requestMap) {
        if ((internalMulti != null) && (internalMulti.isAuthentificationSharingAllowed())) {
            HttpSession session = requestMap.getRequest().getSession(false);
            if (session != null) {
                BDFSession bdfSession = (BDFSession) session.getAttribute(BDFSESSION_KEY);
                if (bdfSession != null) {
                    Redacteur matchingRedacteur = bdfSession.checkMatchingBdfUser(bdfServer);
                    if (matchingRedacteur != null) {
                        BdfUser bdfUser = createBdfUser(bdfServer, bdfSession, matchingRedacteur);
                        checkBdfUser(bdfUser, requestMap, bdfServer.getName());
                        return bdfUser;
                    }
                }
            }
        }
        return null;
    }

    public synchronized void invalidateBdfUser(BdfServer bdfServer, BdfUser bdfUser, boolean all) {
        String name = bdfServer.getName();
        BDFSession userBdfSession = null;
        for (BDFSession bdfSession : sessionsMap.values()) {
            if (bdfSession.containsBdfUser(name, bdfUser)) {
                userBdfSession = bdfSession;
                break;
            }
        }
        if (userBdfSession != null) {
            if (all) {
                userBdfSession.removeRedacteur(bdfUser.getRedacteur());
            } else {
                userBdfSession.removeBdfUser(name);
            }
            if (userBdfSession.isEmpty()) {
                HttpSession httpSession = userBdfSession.getHttpSession();
                sessionsMap.remove(httpSession.getId());
                httpSession.removeAttribute(BDFSESSION_KEY);
                httpSession.removeAttribute(BDFBINDING_KEY);
                httpSession.invalidate();
            }
        }
    }

    public synchronized BdfUser createBdfUser(BdfServer bdfServer, HttpServletRequestMap requestMap) throws LoginException {
        String bdfLogin = requestMap.getParameter(InteractionConstants.BDF_LOGIN_PARAMETER);
        if (bdfLogin == null) {
            bdfLogin = requestMap.getParameter("bdf-idsphere"); //Ancienne version
        }
        if (bdfLogin == null) {
            throw new LoginException(LoginException.UNDEFINED_ERROR);
        }
        String bdfPassword = requestMap.getParameter(InteractionConstants.BDF_PASSWORD_PARAMETER);
        if (bdfPassword == null) {
            throw new LoginException(LoginException.UNDEFINED_ERROR);
        }
        String bdfSphere = requestMap.getParameter(InteractionConstants.BDF_SPHERE_PARAMETER);
        bdfLogin = bdfLogin.trim();
        if ((!bdfLogin.contains("[")) && (bdfSphere != null)) {
            bdfSphere = bdfSphere.trim();
            if (bdfSphere.length() > 0) {
                bdfLogin = bdfLogin + "[" + bdfSphere + "]";
            }
        }
        Redacteur redacteur;
        try {
            redacteur = SphereUtils.parse(bdfServer.getFichotheque(), bdfLogin);
        } catch (SphereUtils.RedacteurLoginException rle) {
            if (rle.getType() == SphereUtils.UNKNOWN_SPHERE) {
                throw new LoginException(LoginException.SPHERE_ERROR);
            } else {
                throw new LoginException(LoginException.LOGIN_ERROR);
            }
        }
        if (redacteur == null) {
            throw new LoginException(LoginException.LOGIN_ERROR);
        }
        if (!redacteur.isActive()) {
            throw new LoginException(LoginException.INACTIF_ERROR);
        }
        if (!loginPasswordDisabled) {
            if (!bdfServer.getPasswordManager().checkPassword(redacteur.getGlobalId(), bdfPassword)) {
                throw new LoginException(LoginException.PASSWORD_ERROR);
            }
        }
        HttpSession httpSession = requestMap.getRequest().getSession();
        httpSession.setMaxInactiveInterval(maxInactiveInterval);
        BDFSession bdfSession = getOrCreateBDFSession(httpSession);
        BdfUser bdfUser = createBdfUser(bdfServer, bdfSession, redacteur);
        checkBdfUser(bdfUser, requestMap, bdfServer.getName());
        return bdfUser;
    }

    public static BDFWebapps build(ServletContext servletContext) {
        String servletContextName = ServletUtils.getServletContextName(servletContext);
        SmtpManager smtpManager = BDFInit.smtpManager(servletContext);
        boolean loginPasswordDisabled = BDFInit.loginPasswordDisabled(servletContext);
        boolean authentificationSharingAllowed = BDFInit.authentificationSharingAllowed(servletContext);
        Conf conf = BDFInit.conf(servletContext);
        return new BDFWebapps(servletContextName, conf, smtpManager, loginPasswordDisabled, authentificationSharingAllowed);
    }

    private BdfUser createBdfUser(BdfServer bdfServer, BDFSession bdfSession, Redacteur redacteur) {
        BdfUser bdfUser = bdfServer.createBdfUser(redacteur);
        bdfSession.addBdfUser(bdfServer.getName(), bdfUser);
        bdfServer.updateFicheSelection(bdfUser);
        return bdfUser;
    }

    private void checkBdfUser(BdfUser bdfUser, HttpServletRequestMap requestMap, String fichothequeName) {
        long currentTime = System.currentTimeMillis();
        boolean withJavascript = true;
        if (requestMap.isTrue(InteractionConstants.BDF_NOSCRIPT_PARAMETER)) {
            withJavascript = false;
        }
        bdfUser.putParameter(BdfUserConstants.SESSION_CREATIONTIME, currentTime);
        bdfUser.putParameter(BdfUserConstants.SESSION_LASTACCESSEDTIME, currentTime);
        bdfUser.putParameter(BdfUserConstants.HTML_WITHJAVASCRIPT, withJavascript);
        bdfUser.putParameter(BdfUserConstants.SESSION_ROOTURL, BDFUtils.getRootUrl(requestMap.getRequest(), fichothequeName));
    }

    private BDFSession getOrCreateBDFSession(HttpSession httpSession) {
        BDFSession bdfSession = (BDFSession) httpSession.getAttribute(BDFSESSION_KEY);
        if (bdfSession == null) {
            bdfSession = new BDFSession(httpSession);
            httpSession.setAttribute(BDFSESSION_KEY, bdfSession);
            sessionsMap.put(httpSession.getId(), bdfSession);
            httpSession.setAttribute(BDFBINDING_KEY, bdfBinding);
        }
        return bdfSession;
    }


    public class Unique {

        private final ResourceStorages webappsResourceStorages;
        private final BuildInfo buildInfo;
        private final SmtpManager smtpManager;
        private final MimeTypeResolver mimeTypeResolver;
        private final List<BdfExtensionInitializer> extensionInitializerList;
        private BdfServer bdfServer = null;

        private Unique(ResourceStorages webappsResourceStorages, BuildInfo buildInfo, SmtpManager smtpManager, MimeTypeResolver mimeTypeResolver, List<BdfExtensionInitializer> extensionInitializerList) {
            this.buildInfo = buildInfo;
            this.webappsResourceStorages = webappsResourceStorages;
            this.smtpManager = smtpManager;
            this.mimeTypeResolver = mimeTypeResolver;
            this.extensionInitializerList = extensionInitializerList;
        }

        public BdfServer getBdfServer() {
            if (bdfServer != null) {
                return bdfServer;
            }
            return initBdfServer();
        }

        private synchronized BdfServer initBdfServer() {
            if (bdfServer != null) {
                return bdfServer;
            }
            String name = ConfConstants.UNIQUE_NAME;
            Map<String, Object> contextObjectMap = new HashMap<String, Object>();
            contextObjectMap.put(BdfServerConstants.SMTPMANAGER_CONTEXTOBJECT, smtpManager);
            contextObjectMap.put(BdfServerConstants.SESSIONOBSERVER_CONTEXTOBJECT, new InternalSessionObserver(name));
            contextObjectMap.put(BdfServerConstants.MIMETYPERESOLVER_CONTEXTOBJECT, mimeTypeResolver);
            Storages storages = BDFInit.storages(conf.getBdfServerDirs(name), webappsResourceStorages);
            bdfServer = new BdfServerImpl(name, storages, buildInfo, extensionInitializerList, contextObjectMap);
            return bdfServer;
        }

    }


    private class InternalMulti implements Multi {

        private final MultiConf multiConf;
        private final String servletContextName;
        private final BdfMessageLocalisationFactory messageLocalisationFactory;
        private final InternalMessageLocalisationProvider messageLocalisationProvider = new InternalMessageLocalisationProvider();
        private final ResourceStorages webappsResourceStorages;
        private final JsAnalyser webappsJsAnalyser;
        private final boolean authentificationSharingAllowed;
        private final Set<String> unmodifiableExistingNameSet;
        private final BuildInfo buildInfo;
        private final FichothequeSharing fichothequeSharing;
        private final SmtpManager smtpManager;
        private final MimeTypeResolver mimeTypeResolver;
        private final List<BdfExtensionInitializer> extensionInitializerList;
        private final SortedMap<String, String> stateMap = new TreeMap<String, String>();
        private final Map<String, BdfServer> bdfServerMap = new HashMap<String, BdfServer>();
        private final ConfDirs multiConfDirs;
        private final DefaultMultiMetadata multiMetadata;

        private InternalMulti(MultiConf multiConf, String servletContextName, Set<String> existingNameSet, ResourceStorages webappsResourceStorages, boolean authentificationSharingAllowed, BuildInfo buildInfo, SmtpManager smtpManager, MimeTypeResolver mimeTypeResolver, List<BdfExtensionInitializer> extensionInitializerList, ConfDirs multiConfDirs) {
            this.multiConf = multiConf;
            this.servletContextName = servletContextName;
            this.messageLocalisationFactory = BdfMessageLocalisationFactory.buildFactory(webappsResourceStorages);
            for (String name : existingNameSet) {
                stateMap.put(name, Multi.NOT_INIT_STATE);
            }
            this.webappsResourceStorages = webappsResourceStorages;
            this.webappsJsAnalyser = new ResourceJsAnalyser(webappsResourceStorages);
            this.authentificationSharingAllowed = authentificationSharingAllowed;
            this.unmodifiableExistingNameSet = Collections.unmodifiableSet(stateMap.keySet());
            this.buildInfo = buildInfo;
            if (authentificationSharingAllowed) {
                this.fichothequeSharing = new InternalFichothequeSharing();
            } else {
                this.fichothequeSharing = null;
            }
            this.smtpManager = smtpManager;
            this.mimeTypeResolver = mimeTypeResolver;
            this.extensionInitializerList = extensionInitializerList;
            this.multiConfDirs = multiConfDirs;
            this.multiMetadata = DefaultMultiMetadata.build(multiConfDirs, multiConf);
            checkMetadata();
        }

        @Override
        public MultiConf getMultiConf() {
            return multiConf;
        }

        @Override
        public String getServletContextName() {
            return servletContextName;
        }

        @Override
        public ResourceStorages getWebappsResourceStorages() {
            return webappsResourceStorages;
        }

        @Override
        public MimeTypeResolver getMimeTypeResolver() {
            return mimeTypeResolver;
        }

        @Override
        public JsAnalyser getWebappsJsAnalyser() {
            return webappsJsAnalyser;
        }

        @Override
        public boolean isAuthentificationSharingAllowed() {
            return authentificationSharingAllowed;
        }

        @Override
        public Set<String> getExistingNameSet() {
            return unmodifiableExistingNameSet;
        }

        @Override
        public BuildInfo getBuildInfo() {
            return buildInfo;
        }

        @Override
        public String getState(String fichothequeName) {
            String state = stateMap.get(fichothequeName);
            if (state != null) {
                return state;
            } else {
                return Multi.NOT_EXISTING_STATE;
            }
        }

        @Override
        public BdfServer getBdfServer(String name) throws ErrorMessageException {
            BdfServer bdfServer = bdfServerMap.get(name);
            if (bdfServer != null) {
                return bdfServer;
            }
            String state = stateMap.get(name);
            if (state == null) {
                boolean here = false;
                if ((multiConf != null) && (ConfUtils.isValidMultiName(name))) {
                    if (ConfUtils.isExistingName(conf, name)) {
                        here = true;
                    }
                }
                if (here) {
                    name = name.intern();
                    stateMap.put(name, Multi.NOT_INIT_STATE);
                } else {
                    throw new ErrorMessageException(MultiConstants.UNKNOWN_FICHOTHEQUE_ERROR, name);
                }
            } else if (state.equals(Multi.INACTIVE_STATE)) {
                throw new ErrorMessageException(MultiConstants.INACTIVE_FICHOTHEQUE_ERROR);
            }
            String intern = name.intern();
            synchronized (intern) {
                bdfServer = bdfServerMap.get(intern);
                if (bdfServer == null) {
                    Map<String, Object> contextObjectMap = new HashMap<String, Object>();
                    contextObjectMap.put(BdfServerConstants.SMTPMANAGER_CONTEXTOBJECT, smtpManager);
                    contextObjectMap.put(BdfServerConstants.SESSIONOBSERVER_CONTEXTOBJECT, new InternalSessionObserver(name));
                    contextObjectMap.put(BdfServerConstants.FICHOTHEQUESHARING_CONTEXTOBJECT, fichothequeSharing);
                    contextObjectMap.put(BdfServerConstants.MIMETYPERESOLVER_CONTEXTOBJECT, mimeTypeResolver);
                    Storages storages = BDFInit.storages(conf.getBdfServerDirs(name), webappsResourceStorages);
                    bdfServer = new BdfServerImpl(name, storages, buildInfo, extensionInitializerList, contextObjectMap);
                    bdfServerMap.put(intern, bdfServer);
                    stateMap.put(intern, Multi.ACTIVE_STATE);
                }
            }
            return bdfServer;
        }

        @Override
        public BdfServerDirs getBdfServerDirs(String name) {
            if (multiConf == null) {
                name = ConfConstants.UNIQUE_NAME;
            }
            if (!stateMap.containsKey(name)) {
                return null;
            }
            return conf.getBdfServerDirs(name);
        }

        @Override
        public ConfDirs getMultiDirs() {
            return multiConfDirs;
        }

        @Override
        public MultiMetadata getMultiMetadata() {
            return multiMetadata;
        }

        @Override
        public MultiEditor getMultiEditor(EditOrigin editOrigin) {
            return new InternalMultiEditor(editOrigin);
        }

        @Override
        public MessageLocalisationProvider getMessageLocalisationProvider() {
            return messageLocalisationProvider;
        }

        @Override
        public synchronized void createFichotheque(CreationParameters creationParameters) {
            String newFichothequeName = creationParameters.getNewFichothequeName();
            if (!ConfUtils.isValidMultiName(newFichothequeName)) {
                throw new IllegalArgumentException("wrong name: " + newFichothequeName);
            }
            if (stateMap.containsKey(newFichothequeName)) {
                throw new IllegalArgumentException("existing name: " + newFichothequeName);
            }
            newFichothequeName = newFichothequeName.intern();
            BdfServerDirs dirs = conf.getBdfServerDirs(newFichothequeName);
            switch (creationParameters.getType()) {
                case CreationParameters.NEW_TYPE:
                    BdfdataDirectories.writeStartFiles(dirs, creationParameters.startValues(), true);
                    break;
                case CreationParameters.NEW_WITH_SPHERES_TYPE:
                    BdfdataDirectories.writeStartFiles(dirs, creationParameters.startValues(), false);
                    BdfdataDirectories bdfdataDirectories = BdfdataDirectories.build(dirs);
                    SphereCopyEngine.run(bdfdataDirectories, creationParameters.getSourceBdfServer());
                    break;
                case CreationParameters.DUPLICATE_TYPE:
                try {
                    File tempZipFile = File.createTempFile("bdf", "zip");
                    BdfServerZip bdfServerZip = new BdfServerZip(creationParameters.getSourceBdfServer(), creationParameters.getSubsetEligibility());
                    try (ZipOutputStream os = new ZipOutputStream(new FileOutputStream(tempZipFile))) {
                        os.setLevel(Deflater.NO_COMPRESSION);
                        bdfServerZip.zip(os);
                    }
                    ConfUtils.copyZippedFiles(dirs, tempZipFile);
                    tempZipFile.delete();
                } catch (IOException ioe) {
                    throw new NestedIOException(ioe);
                }
                break;
            }
            stateMap.put(newFichothequeName, Multi.ACTIVE_STATE);
        }

        private void checkMetadata() {
            Attribute inactiveAttribute = multiMetadata.getAttributes().getAttribute(FichothequeSpace.INACTIVE_KEY);
            Set<String> newInactiveSet = new HashSet<String>();
            if (inactiveAttribute != null) {
                for (String value : inactiveAttribute) {
                    newInactiveSet.add(value);
                }
            }
            for (Map.Entry<String, String> entry : stateMap.entrySet()) {
                if (entry.getValue().equals(Multi.INACTIVE_STATE)) {
                    if (!newInactiveSet.contains(entry.getKey())) {
                        entry.setValue(Multi.NOT_INIT_STATE);
                    }
                }
            }
            for (String name : newInactiveSet) {
                if (stateMap.containsKey(name)) {
                    stateMap.put(name, Multi.INACTIVE_STATE);
                    bdfServerMap.remove(name);
                }
            }
        }


        private class InternalFichothequeSharing implements FichothequeSharing {


            private InternalFichothequeSharing() {

            }

            @Override
            public Fichotheque getFichotheque(String name) {
                try {
                    return getBdfServer(name).getFichotheque();
                } catch (ErrorMessageException eme) {
                    return null;
                }
            }

        }


        private class InternalMessageLocalisationProvider implements MessageLocalisationProvider {

            private InternalMessageLocalisationProvider() {

            }

            @Override
            public MessageLocalisation getMessageLocalisation(LangPreference langPreference, Locale formatLocale) {
                return messageLocalisationFactory.newInstance(langPreference, formatLocale);
            }

            @Override
            public MessageLocalisation getMessageLocalisation(Lang lang) {
                return messageLocalisationFactory.newInstance(lang);
            }

        }


        private class InternalMultiEditor implements MultiEditor {

            private final EditOrigin editOrigin;
            private MultiMetadataEditor multiMetadataEditor = null;

            private InternalMultiEditor(EditOrigin editOrigin) {
                this.editOrigin = editOrigin;
            }

            @Override
            public Multi getMulti() {
                return InternalMulti.this;
            }

            @Override
            public EditOrigin getEditOrigin() {
                return editOrigin;
            }

            @Override
            public MultiMetadataEditor getMultiMetadataEditor() {
                if (multiMetadataEditor == null) {
                    multiMetadataEditor = multiMetadata.getMultiMetadataEditor(editOrigin);
                }
                return multiMetadataEditor;
            }

            @Override
            public void saveChanges() {
                if (multiMetadataEditor != null) {
                    multiMetadata.saveChanges(multiMetadataEditor);
                    checkMetadata();
                }
            }

        }

    }


    private class BDFBinding implements HttpSessionBindingListener {

        private BDFBinding() {

        }

        @Override
        public void valueBound(HttpSessionBindingEvent httpSessionEvent) {
        }

        @Override
        public void valueUnbound(HttpSessionBindingEvent httpSessionEvent) {
            HttpSession session = httpSessionEvent.getSession();
            sessionsMap.remove(session.getId());
        }

    }


    private class InternalSessionObserver implements SessionObserver {

        private final String fichothequeName;

        private InternalSessionObserver(String fichothequeName) {
            this.fichothequeName = fichothequeName;
        }

        @Override
        public synchronized BdfUser[] getCurrentBdfUserArray() {
            List<BdfUser> bdfUserList = new ArrayList<BdfUser>();
            for (BDFSession bdfSession : sessionsMap.values()) {
                BdfUser bdfUser = bdfSession.getBdfUser(fichothequeName);
                if (bdfUser != null) {
                    bdfUserList.add(bdfUser);
                }
            }
            return bdfUserList.toArray(new BdfUser[bdfUserList.size()]);
        }

    }


}
