/* BdfServer - Copyright (c) 2012-2022 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.tools.exportation.table;

import fr.exemole.bdfserver.api.BdfServer;
import fr.exemole.bdfserver.api.exportation.table.FicheTableParameters;
import fr.exemole.bdfserver.api.roles.PermissionSummary;
import fr.exemole.bdfserver.api.ui.FieldUi;
import fr.exemole.bdfserver.api.ui.SpecialIncludeUi;
import fr.exemole.bdfserver.api.ui.SubsetIncludeUi;
import fr.exemole.bdfserver.api.ui.UiComponent;
import fr.exemole.bdfserver.api.ui.UiComponents;
import java.util.ArrayList;
import java.util.List;
import net.fichotheque.FichothequeConstants;
import net.fichotheque.Subset;
import net.fichotheque.SubsetKey;
import net.fichotheque.corpus.Corpus;
import net.fichotheque.corpus.metadata.CorpusField;
import net.fichotheque.corpus.metadata.CorpusMetadata;
import net.fichotheque.corpus.metadata.FieldKey;
import net.fichotheque.exportation.table.FormatColDef;
import net.fichotheque.exportation.table.TableDef;
import net.fichotheque.exportation.table.TableDefItem;
import net.fichotheque.format.FormatSourceKey;
import net.fichotheque.format.SubsetPathKey;
import net.fichotheque.include.ExtendedIncludeKey;
import net.fichotheque.thesaurus.Thesaurus;
import net.fichotheque.tools.format.FichothequeFormatDefBuilder;
import net.fichotheque.utils.TableDefUtils;
import net.mapeadores.util.format.FormatConstants;


/**
 *
 * @author Vincent Calame
 */
public class FicheTableDefBuilder {

    private final BdfServer bdfServer;
    private final Corpus corpus;
    private final CorpusMetadata corpusMetadata;
    private final FicheTableParameters ficheTableParameters;
    private final PermissionSummary permissionSummary;
    private final SubsetPathKey masterPath;
    private final List<TableDefItem> tableDefItemList = new ArrayList<TableDefItem>();

    private FicheTableDefBuilder(BdfServer bdfServer, Corpus corpus, FicheTableParameters ficheTableParameters, PermissionSummary permissionSummary) {
        this.bdfServer = bdfServer;
        this.corpus = corpus;
        this.permissionSummary = permissionSummary;
        this.corpusMetadata = corpus.getCorpusMetadata();
        this.ficheTableParameters = ficheTableParameters;
        Subset masterSubset = corpus.getMasterSubset();
        if ((masterSubset != null) && (masterSubset instanceof Thesaurus) && (permissionSummary.hasAccess(masterSubset))) {
            masterPath = SubsetPathKey.newRattachementPath(masterSubset.getSubsetKey());
        } else {
            masterPath = null;
        }
    }

    public TableDef toTableDef() {
        return TableDefUtils.toTableDef(tableDefItemList);
    }

    public static TableDef fromUi(BdfServer bdfServer, Corpus corpus, FicheTableParameters ficheTableParameters, PermissionSummary permissionSummary) {
        FicheTableDefBuilder builder = new FicheTableDefBuilder(bdfServer, corpus, ficheTableParameters, permissionSummary);
        builder.init();
        return builder.toTableDef();
    }

    private void init() {
        tableDefItemList.add(getCorpusFieldColDef(FieldKey.ID));
        if (ficheTableParameters.isWith(FicheTableParameters.WITH_CHRONO)) {
            tableDefItemList.add(getChronoColDef(FichothequeConstants.DATECREATION_NAME));
            tableDefItemList.add(getChronoColDef(FichothequeConstants.DATEMODIFICATION_NAME));
        }
        UiComponents uiComponents = bdfServer.getUiManager().getMainUiComponents(corpus);
        for (UiComponent uiComponent : uiComponents.getUiComponentList()) {
            if (uiComponent instanceof FieldUi) {
                FieldUi fieldUi = (FieldUi) uiComponent;
                FieldKey fieldKey = fieldUi.getFieldKey();
                if ((!ficheTableParameters.isWith(FicheTableParameters.WITH_SECTION)) && (fieldKey.isSection())) {
                    continue;
                }
                CorpusField corpusField = corpusMetadata.getCorpusField(fieldKey);
                if (corpusField != null) {
                    tableDefItemList.add(FicheTableDefBuilder.this.getCorpusFieldColDef(corpusField));
                }
            } else if (uiComponent instanceof SpecialIncludeUi) {
                FormatColDef formatColDef = getSpecialIncludeColDef(uiComponent.getName());
                if (formatColDef != null) {
                    tableDefItemList.add(formatColDef);
                }
            } else if (uiComponent instanceof SubsetIncludeUi) {
                SubsetIncludeUi includeUi = (SubsetIncludeUi) uiComponent;
                ExtendedIncludeKey extendedIncludeKey = includeUi.getExtendedIncludeKey();
                if (testInclude(extendedIncludeKey)) {
                    Subset otherSubset = bdfServer.getFichotheque().getSubset(extendedIncludeKey.getSubsetKey());
                    if ((otherSubset != null) && (permissionSummary.hasAccess(otherSubset))) {
                        tableDefItemList.add(getSubsetIncludeColDef(extendedIncludeKey));
                    }
                }
            }
        }
    }

    private boolean testInclude(ExtendedIncludeKey extendedIncludeKey) {
        switch (extendedIncludeKey.getCategory()) {
            case SubsetKey.CATEGORY_ADDENDA:
                return ficheTableParameters.isWith(FicheTableParameters.WITH_ADDENDAINCLUDE);
            case SubsetKey.CATEGORY_ALBUM:
                return ficheTableParameters.isWith(FicheTableParameters.WITH_ALBUMINCLUDE);
            case SubsetKey.CATEGORY_CORPUS:
                return ficheTableParameters.isWith(FicheTableParameters.WITH_CORPUSINCLUDE);
            case SubsetKey.CATEGORY_THESAURUS:
                return ficheTableParameters.isWith(FicheTableParameters.WITH_THESAURUSINCLUDE);
            default:
                return false;
        }
    }

    private FormatColDef getCorpusFieldColDef(FieldKey fieldKey) {
        return FicheTableDefBuilder.this.getCorpusFieldColDef(corpusMetadata.getCorpusField(fieldKey));
    }

    private FormatColDef getCorpusFieldColDef(CorpusField corpusField) {
        short patternMode = ficheTableParameters.getPatternMode();
        FichothequeFormatDefBuilder formatDefBuilder = new FichothequeFormatDefBuilder();
        formatDefBuilder.addSource(FormatSourceKey.newFieldKeyInstance(corpusField.getFieldKey()));
        if (isWithIntegerCast(corpusField)) {
            formatDefBuilder.setCastType(FormatConstants.INTEGER_CAST);
        }
        switch (patternMode) {
            case FicheTableParameters.FORMSYNTAX_PATTERNMODE: {
                formatDefBuilder.addFormatPattern("{formsyntax}");
                break;
            }
            case FicheTableParameters.STANDARD_PATTERNMODE: {
                if (isWithSeparator(corpusField)) {
                    formatDefBuilder.putStringValue(FormatConstants.SEPARATOR_PARAMKEY, ", ");
                }
                switch (corpusField.getFicheItemType()) {
                    case CorpusField.MONTANT_FIELD:
                        formatDefBuilder.addFormatPattern("{code}");
                        formatDefBuilder.setCastType(FormatConstants.MONEY_CAST);
                        break;
                    case CorpusField.NOMBRE_FIELD:
                        formatDefBuilder.addFormatPattern("{label}");
                        break;
                }
                break;
            }
            case FicheTableParameters.CODE_PATTERNMODE: {
                switch (corpusField.getFicheItemType()) {
                    case CorpusField.MONTANT_FIELD:
                    case CorpusField.NOMBRE_FIELD:
                    case CorpusField.PAYS_FIELD:
                    case CorpusField.LANGUE_FIELD:
                        formatDefBuilder.addFormatPattern("{code}");
                        break;
                }
                break;
            }
        }
        return TableDefUtils.toFormatColDef(corpusField.getFieldString(), formatDefBuilder.toFichothequeFormatDef());
    }

    private FormatColDef getSpecialIncludeColDef(String specialIncludeName) {
        if (!specialIncludeName.equals(FichothequeConstants.LIAGE_NAME)) {
            return null;
        }
        if (!ficheTableParameters.isWith(FicheTableParameters.WITH_CORPUSINCLUDE)) {
            return null;
        }
        FichothequeFormatDefBuilder formatDefBuilder = new FichothequeFormatDefBuilder();
        formatDefBuilder.addSource(FormatSourceKey.newSpecialIncludeNameInstance(specialIncludeName));
        switch (ficheTableParameters.getPatternMode()) {
            case FicheTableParameters.STANDARD_PATTERNMODE:
                formatDefBuilder.addFormatPattern("{titre}");
                formatDefBuilder.putStringValue(FormatConstants.SEPARATOR_PARAMKEY, ", ");
                break;
            case FicheTableParameters.CODE_PATTERNMODE:
                formatDefBuilder.addFormatPattern("{corpus}/{id}");
                break;
            case FicheTableParameters.FORMSYNTAX_PATTERNMODE:
                formatDefBuilder.addFormatPattern("{formsyntax}");
                break;
        }
        return TableDefUtils.toFormatColDef(specialIncludeName, formatDefBuilder.toFichothequeFormatDef());
    }

    private FormatColDef getSubsetIncludeColDef(ExtendedIncludeKey includeKey) {
        FichothequeFormatDefBuilder formatDefBuilder = new FichothequeFormatDefBuilder();
        if ((includeKey.isMaster()) && (masterPath != null)) {
            formatDefBuilder.addSource(FormatSourceKey.newIncludeKeyInstance(includeKey.getRootIncludeKey(), masterPath));
        } else {
            formatDefBuilder.addSource(FormatSourceKey.newIncludeKeyInstance(includeKey.getRootIncludeKey()));
        }
        SubsetKey subsetKey = includeKey.getSubsetKey();
        if (subsetKey.isThesaurusSubset()) {
            boolean isIdalpha = isIdalpha(subsetKey);
            switch (ficheTableParameters.getPatternMode()) {
                case FicheTableParameters.STANDARD_PATTERNMODE:
                    if (isIdalpha) {
                        formatDefBuilder.addFormatPattern("{idalpha} – {label}");
                    } else {
                        formatDefBuilder.addFormatPattern("{label}");
                    }
                    break;
                case FicheTableParameters.CODE_PATTERNMODE:
                    if (isIdalpha) {
                        formatDefBuilder.addFormatPattern("{idalpha}");
                    } else {
                        formatDefBuilder.addFormatPattern("{id}");
                    }
                    break;
                case FicheTableParameters.FORMSYNTAX_PATTERNMODE:
                    formatDefBuilder.addFormatPattern("{formsyntax}");
                    break;
            }
        } else if (subsetKey.isCorpusSubset()) {
            switch (ficheTableParameters.getPatternMode()) {
                case FicheTableParameters.STANDARD_PATTERNMODE:
                    formatDefBuilder.addFormatPattern("{titre}");
                    formatDefBuilder.putStringValue(FormatConstants.SEPARATOR_PARAMKEY, ", ");
                    break;
                case FicheTableParameters.CODE_PATTERNMODE:
                    formatDefBuilder.addFormatPattern("{id}");
                    break;
                case FicheTableParameters.FORMSYNTAX_PATTERNMODE:
                    formatDefBuilder.addFormatPattern("{formsyntax}");
                    break;
            }
        } else {
            switch (ficheTableParameters.getPatternMode()) {
                case FicheTableParameters.STANDARD_PATTERNMODE:
                case FicheTableParameters.CODE_PATTERNMODE:
                    break;
                case FicheTableParameters.FORMSYNTAX_PATTERNMODE:
                    formatDefBuilder.addFormatPattern("{formsyntax}");
                    break;
            }
        }
        return TableDefUtils.toFormatColDef(includeKey.toString(), formatDefBuilder.toFichothequeFormatDef());
    }

    private FormatColDef getChronoColDef(String name) {
        FichothequeFormatDefBuilder formatDefBuilder = new FichothequeFormatDefBuilder();
        formatDefBuilder.addSource(FormatSourceKey.newSpecialIncludeNameInstance(name));
        switch (ficheTableParameters.getPatternMode()) {
            case FicheTableParameters.STANDARD_PATTERNMODE:
            case FicheTableParameters.CODE_PATTERNMODE:
                formatDefBuilder.addFormatPattern("{iso}");
                formatDefBuilder.setCastType(FormatConstants.DATE_CAST);
                break;
            case FicheTableParameters.FORMSYNTAX_PATTERNMODE:
                formatDefBuilder.addFormatPattern("{formsyntax}");
                break;
        }
        return TableDefUtils.toFormatColDef(name, formatDefBuilder.toFichothequeFormatDef());
    }

    private boolean isIdalpha(SubsetKey subsetKey) {
        Thesaurus thesaurus = (Thesaurus) corpus.getFichotheque().getSubset(subsetKey);
        if (thesaurus != null) {
            return thesaurus.isIdalphaType();
        } else {
            return false;
        }
    }

    private static boolean isWithSeparator(CorpusField corpusField) {
        switch (corpusField.getCategory()) {
            case FieldKey.INFORMATION_CATEGORY:
                return true;
            case FieldKey.SPECIAL_CATEGORY:
                switch (corpusField.getFieldString()) {
                    case FieldKey.SPECIAL_REDACTEURS:
                        return true;

                }
            default:
                return false;
        }
    }

    private static boolean isWithIntegerCast(CorpusField corpusField) {
        switch (corpusField.getFieldString()) {
            case FieldKey.SPECIAL_ID:
                return true;
            default:
                return false;

        }
    }

}
