/* BdfServer_Multi - Copyright (c) 2022 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.multi.tools;

import fr.exemole.bdfserver.api.storage.BdfStorageException;
import fr.exemole.bdfserver.conf.ConfDirs;
import fr.exemole.bdfserver.multi.MultiConf;
import fr.exemole.bdfserver.multi.api.MultiMetadata;
import fr.exemole.bdfserver.multi.api.MultiMetadataEditor;
import fr.exemole.bdfserver.storage.directory.bdfdata.BdfStorageUtils;
import fr.exemole.bdfserver.storage.directory.bdfdata.StorageDirectories;
import fr.exemole.bdfserver.storage.directory.bdfdata.StorageFile;
import java.io.BufferedWriter;
import java.io.IOException;
import java.text.ParseException;
import net.fichotheque.EditOrigin;
import net.fichotheque.Metadata;
import net.fichotheque.utils.DefaultMetadata;
import net.mapeadores.util.attr.Attribute;
import net.mapeadores.util.attr.AttributeKey;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.Langs;
import net.mapeadores.util.localisation.LangsUtils;
import net.mapeadores.util.logging.LogUtils;
import net.mapeadores.util.text.Label;
import net.mapeadores.util.text.StringUtils;
import net.mapeadores.util.xml.AppendableXMLWriter;
import net.mapeadores.util.xml.XMLUtils;
import org.w3c.dom.Document;


/**
 *
 * @author Vincent Calame
 */
public class DefaultMultiMetadata extends DefaultMetadata implements MultiMetadata {

    private final StorageDirectories storageDirectories;
    private Langs workingLangs;
    private String authority;

    public DefaultMultiMetadata(StorageDirectories storageDirectories, Langs workingLangs, String authority) {
        this.storageDirectories = storageDirectories;
        this.workingLangs = workingLangs;
        this.authority = authority;
    }

    @Override
    public Langs getWorkingLangs() {
        return workingLangs;
    }

    @Override
    public String getAuthority() {
        return authority;
    }

    public MultiMetadataEditor getMultiMetadataEditor(EditOrigin editOrigin) {
        return new InternalMultiMetadataEditor(editOrigin);
    }

    public boolean saveChanges(MultiMetadataEditor editor) {
        InternalMultiMetadataEditor internalEditor = (InternalMultiMetadataEditor) editor;
        return internalEditor.saveChanges();
    }

    private StorageFile getMetadataStorageFile() {
        return getMetadataStorageFile(storageDirectories);
    }

    public static StorageFile getMetadataStorageFile(StorageDirectories storageDirectories) {
        String path = "metadata.xml";
        return StorageFile.build(storageDirectories, path);
    }

    public static DefaultMultiMetadata build(ConfDirs multiConfDirs, MultiConf multiConf) {
        StorageDirectories storageDirectories = BdfStorageUtils.toStorageDirectories(multiConfDirs);
        DefaultMultiMetadata multiMetadata = new DefaultMultiMetadata(storageDirectories, LangsUtils.wrap(getDefaultLang(multiConf)), getDefaultAuthority(multiConf));
        StorageFile storageFile = getMetadataStorageFile(storageDirectories);
        if (storageFile.exists()) {
            Document document = storageFile.readDocument();
            MultiMetadataDOMReader reader = new MultiMetadataDOMReader(multiMetadata.getMultiMetadataEditor(null), LogUtils.NULL_MULTIMESSAGEHANDLER);
            reader.read(document.getDocumentElement());
        }
        return multiMetadata;
    }

    private static String getDefaultAuthority(MultiConf multiConf) {
        String authority = multiConf.getParam("authority");
        if (authority != null) {
            if (!StringUtils.isAuthority(authority)) {
                authority = null;
            }
        }
        if (authority == null) {
            authority = "localhost";
        }
        return authority;
    }

    private static Lang getDefaultLang(MultiConf multiConf) {
        Lang defaultLang = null;
        String langString = multiConf.getParam("lang");
        if (langString != null) {
            try {
                defaultLang = Lang.parse(langString);
            } catch (ParseException pe) {
            }
        }
        if (defaultLang == null) {
            defaultLang = Lang.getDefault().getRootLang();
        }
        return defaultLang;
    }


    private class InternalMultiMetadataEditor implements MultiMetadataEditor {

        private final EditOrigin editOrigin;
        private boolean withChanges = false;

        private InternalMultiMetadataEditor(EditOrigin editOrigin) {
            this.editOrigin = editOrigin;
        }

        @Override
        public boolean setWorkingLangs(Langs workingLangs) {
            if (workingLangs.isEmpty()) {
                return false;
            } else if (LangsUtils.areEquals(DefaultMultiMetadata.this.workingLangs, workingLangs)) {
                return false;
            } else {
                DefaultMultiMetadata.this.workingLangs = workingLangs;
                withChanges = true;
                return true;
            }
        }

        @Override
        public boolean setAuthority(String authority) {
            if (StringUtils.isAuthority(authority)) {
                String currentAuthority = DefaultMultiMetadata.this.authority;
                if (!currentAuthority.equals(authority)) {
                    DefaultMultiMetadata.this.authority = authority;
                    withChanges = true;
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }

        @Override
        public Metadata getMetadata() {
            return DefaultMultiMetadata.this;
        }

        @Override
        public boolean putLabel(String name, Label label) {
            boolean done = DefaultMultiMetadata.this.putLabel(name, label);
            if (done) {
                withChanges = true;
            }
            return done;
        }

        @Override
        public boolean removeLabel(String name, Lang lang) {
            boolean done = DefaultMultiMetadata.this.removeLabel(name, lang);
            if (done) {
                withChanges = true;
            }
            return done;
        }

        @Override
        public boolean removeAttribute(AttributeKey attributeKey) {
            boolean done = DefaultMultiMetadata.this.removeAttribute(attributeKey);
            if (done) {
                withChanges = true;
            }
            return done;
        }

        @Override
        public boolean putAttribute(Attribute attribute) {
            boolean done = DefaultMultiMetadata.this.putAttribute(attribute);
            if (done) {
                withChanges = true;
            }
            return done;
        }

        private boolean saveChanges() {
            if (withChanges) {
                StorageFile storageFile = getMetadataStorageFile();
                try (BufferedWriter buf = storageFile.archiveAndGetWriter(editOrigin)) {
                    AppendableXMLWriter xmlWriter = XMLUtils.toXMLWriter(buf);
                    xmlWriter.appendXMLDeclaration();
                    MultiMetadataXMLPart multiMetadataXMLPart = new MultiMetadataXMLPart(xmlWriter);
                    multiMetadataXMLPart.appendMultiMetadata(DefaultMultiMetadata.this);
                    return true;
                } catch (IOException ioe) {
                    throw new BdfStorageException(ioe);
                }
            } else {
                return false;
            }
        }

    }

}
